

#include <stdint.h>
#include "Ana.H"


using namespace std;
using namespace avelino;


#define  DISPCNT         *((volatile uint16_t *) 0x04000000)
#define  DISPSTAT        *((volatile uint16_t *) 0x04000004)
#define  MODE3_VRAM_FB    ((uint16_t *) 0x06000000)
#define  OBJ_TILES_VRAM   ((volatile uint16_t *) 0x06014000)
#define  OAM              ((volatile uint16_t *) 0x07000000)
#define  OBJ_PALETTE      ((volatile uint16_t *) 0x05000200)


typedef void (*fptr)(void);


#define  ISR_PTR  *((fptr *) 0x03007FFC)
#define  IME      *((volatile uint32_t *) 0x04000208)
#define  IE       *((volatile uint16_t *) 0x04000200)
#define  IF       *((volatile uint16_t *) 0x04000202)
#define  IFBIOS   *((volatile uint16_t *) 0x03007FF8)


int32_t ballX, ballY, ballDX, ballDY;


void isr()  __attribute__((target("arm")));


void isr() {
	// calculate next location
	ballX += ballDX;
	ballY += ballDY;
	// check collisions
	if (ballX < 0) {
		ballX = 0;
		ballDX = 1;
	}
	else if (ballX > (240 - 8)) {
		ballX = 240 - 8;
		ballDX = -1;
	}
	if (ballY < 0) {
		ballY = 0;
		ballDY = 1;
	}
	else if (ballY > (160 - 8)) {
		ballY = 160 - 8;
		ballDY = -1;
	}
	// locate sprite
	OAM[0] = (ballY & 0x00FF) | ((uint16_t) 1) << 13;
	OAM[1] = (ballX & 0x01FF);
	// mark interrupt as served
	IF = 1;
	IFBIOS |= 1;
}


void enableInterrupts() {
	IME = 0;
	ISR_PTR = isr;
	DISPSTAT = ((uint16_t) 1) << 3;
	IE = 1;   // v-blank
	IME = 1;
}


void memcpy16(uint16_t *dest, const uint16_t *source, uint32_t size) {
	uint16_t *s = (uint16_t *) source;
	while (size > 0) {
		*dest = *s;
		s++;
		dest++;
		size--;
	}
}


const uint16_t BALL_TILE[32] = {
	0x0000, 0x0101, 0x0101, 0x0000,
	0x0100, 0x0202, 0x0202, 0x0001,
	0x0201, 0x0202, 0x0202, 0x0102,
	0x0201, 0x0202, 0x0202, 0x0102,
	0x0201, 0x0202, 0x0202, 0x0102,
	0x0201, 0x0202, 0x0202, 0x0102,
	0x0100, 0x0202, 0x0202, 0x0001,
	0x0000, 0x0101, 0x0101, 0x0000
};


int main() {
	enableInterrupts();
	// force blank to access VRAM
	DISPCNT = ((uint16_t) 1) << 7;
	// draw on framebuffer
	memcpy16(MODE3_VRAM_FB, ANA, 240 * 160);
	// configure OBJ 256 color palette (all colors white)
	OBJ_PALETTE[0] = 0;       // transparent (color don't care)
	OBJ_PALETTE[1] = 0;       // black
	OBJ_PALETTE[2] = 0x7FFF;  // white
	// OBJ tile 0
	memcpy16((uint16_t *) OBJ_TILES_VRAM, BALL_TILE, 32);  // MUST copy 16 bit words
	// configure OBJ 0
	ballX = 120 - 4;
	ballY = 80 - 4;
	//     (y, no rotation/scaling, enabled, no mosaic, 256 colors with 1 palette, shape = square)
	OAM[0] = (ballY & 0x00FF) | ((uint16_t) 1) << 13;
	//     (x, no horizontal flip, no vertical flip, size = 8x8)
	OAM[1] = (ballX & 0x01FF);
	//     (tile number = 512 (first obj tile with mode 3), priority = 0)
	OAM[2] = 512;
	// enable mode 3 with background layer 2, enable obj, enable obj 1-d mapping
	DISPCNT = ((uint16_t) 3) | (((uint16_t) 1) << 10) | (((uint16_t) 1) << 12) | (((uint16_t) 1) << 6);
	ballDX = 1;
	ballDY = 1;
	while (true) {
		asm volatile ("swi 0x05");    // wait for v-blank (armv4 instruction set has no "wfi" instruction)
	}
}
